/***********************************************************************
|  Enhanced C library
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

#include "eclib.h"

#define N_STREAMS 10 /* max nesting level of streams */
#define MAX_PATHNAME 80

extern short ZC_lineno, ZC_pos, ZC_previous_pos;
extern char *ZC_filename;
extern FILE *ZC_file;

int ZC_lastchar;

/* I/O streams, stack of file descriptors, like nested #includes */
static struct {
/* here we have the all too familiar dilema of: reserve space
 * for filenames at compile time (lots of space to be safe,
 * somewhat wasteful in small memory model),
 * or allocate space at run time,
 * or point to the file name and force the calling program
 * to reserve space or allocate it.
 * This is suppose to be an easy-to-use system, so I eschew the last option.
 * Being lazy, I refuse to deal with malloc() here,
 * so I reserve 80 characters per file name.
 * If you insist on having longer path names,  tough!! */
char name[MAX_PATHNAME]; /* name of file being read */
FILE *f; /* file descriptor */
int lastchar; /* I implement my own ungetc, works better with EOF */
short lineno; /* line number */
short pos; /* position in the line */
long ok_to_here; /* offset of last valid position */
} stream_stack[N_STREAMS],
   *cs = 0; /* current stream being read */

io_get()
{
int c;

if(ZC_lastchar) {
c = ZC_lastchar;
ZC_lastchar = 0;
} else {
c = getc(ZC_file);
++ZC_pos;
if(c == '\n') ++ZC_lineno, ZC_previous_pos = ZC_pos, ZC_pos = 0;
}
return c;
} /* io_get */

io_unget(c)
int c;
{
ZC_lastchar = c;
} /* io_unget */

pushstream(name)
char *name;
{
FILE *f;

if(strlen(name) >= MAX_PATHNAME) {
eclib_error(SCAN_LONGNAME, MAX_PATHNAME);
return SCAN_LONGNAME;
}

if(!(f = efopen(name, "r")))
/* error message already printed by efopen */
return SCAN_FILE_ACCESS;

if(cs) { /* recordcurrent location */
if(cs == stream_stack + N_STREAMS - 1) { /* no room */
fclose(f);
eclib_error(SCAN_STREAM_LIMIT, N_STREAMS);
return SCAN_STREAM_LIMIT;
}
io_set_good_point();
++cs;
} else cs = stream_stack;
ZC_file = cs->f = f;
strcpy(cs->name, name);
ZC_filename = cs->name;
ZC_lineno = 1;
ZC_pos = 0;
ZC_lastchar = 0;
io_set_good_point();
return 0; /* ok */
} /* pushstream */

io_set_good_point()
{
cs->ok_to_here = ftell(cs->f);
cs->lineno = ZC_lineno;
cs->pos = ZC_pos;
cs->lastchar = ZC_lastchar;
} /* io_set_good_point */

io_back_good_point()
{
if(fseek(cs->f, cs->ok_to_here, 0))
eclib_error(SCAN_FSEEK, 0);
ZC_lineno = cs->lineno;
ZC_pos = cs->pos;
ZC_lastchar = cs->lastchar;
} /* io_back_good_point */

/* set this flag when parsing a block that should not be split across files */
char ZC_eof_not_allowed;

/* directives to divert input elsewhere */
extern char ZC_divertchars[][4];
extern char ZC_divertnest[];
extern FILE **ZC_divertfiles[];

/* is newline significant, a token in the input language? */
extern char ZC_nl_significant;

int io_skipwhite()
{
int c;
char close_ready, open_ready;
short divnum, nestlev;
char *diversion;
FILE *div_f;

if(ZC_file == 0) goto popstream;

top:
c = io_get();
if(c == '\n' && ZC_nl_significant) return c;
if(isspace(c)) goto top;

/* end of file, pop stream stack */
if(c == EOF) {
if(ZC_eof_not_allowed)
eclib_error(SCAN_EOF_CONTEXT, 0);
fclose(ZC_file);
ZC_file = 0;
popstream:
ZC_filename = 0;
if(!cs ||cs == stream_stack) { cs = 0; return EOF; }
/* pop values */
--cs;
ZC_file = cs->f;
ZC_filename = cs->name;
ZC_lineno = cs->lineno;
ZC_pos = cs->pos;
ZC_lastchar = cs->lastchar;
goto top;
} /* EOF handled */

/* screen for comments or diversions */
for(divnum=0; *(diversion = ZC_divertchars[divnum]); ++divnum) {
if(c != *diversion) continue;

div_f = 0;
if(ZC_divertfiles[divnum])
div_f = *ZC_divertfiles[divnum];
/* div_f may still be null, if the referenced file is closed */

if(diversion[1]) { /* two-char comment */
c = io_get();
if(c != diversion[1]) { /* second char is not correct */
io_unget(c);
c = *diversion;
continue;
}
}

/* ok, we have a valid comment or diversion here */
close_ready = open_ready = 0;
nestlev = 1;
/* no closing char means newline closes the comment */
if(!diversion[2]) diversion[2] = '\n';

while(1) { /* read until the end of the comment */
if(nestlev > 1 && !ZC_divertnest[divnum])
eclib_error(SCAN_NESTED_CMT, 0);

c = io_get();
if(c == EOF)
eclib_error(SCAN_EOF_CMT, 0);

/* look for close comment */
if(c == diversion[2]) {
if(diversion[3]) {
c = io_get();
if(c != diversion[3]) {
if(div_f) putc(diversion[2], div_f);
io_unget(c);
continue;
}
if(!--nestlev) break;
if(div_f) putc(diversion[2], div_f);
if(div_f) putc(c, div_f);
continue;
}
if(!--nestlev) break;
if(div_f) putc(c, div_f);
continue;
}

/* not a close, try for open */
if(div_f) putc(c, div_f);
if(c != diversion[0]) continue;
if(!diversion[1]) { ++nestlev; continue; }
c = io_get();
if(c != diversion[1]) { io_unget(c); continue; }
++nestlev;
} /* end reading comment */

/* send newline to the diversion file whenever a comment ends */
if(div_f) putc('\n', div_f);
/* if line oriented comment and newline is significant, return line break */
if(c == '\n' && ZC_nl_significant) return c;
goto top;
} /* end loop over diversions and comments */

return c;
} /* io_skipwhite */

