/***********************************************************************
|  Enhanced C library
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* mmx.c: very simplistic DOS version of mmx */

/* With all the word processors around, the what-you-see-is-what-you-get
 * programs and the text macro languages, why reinvent any text processing?
 * If all my programs ran under Unix alone, there would indeed be no need.
 * But many of them run under Dos as well,
 * an operating system that lacks a standard word processing package.
 * Nor do I want to require the user to have Word Perfect, or some other
 * ultra-expensive, unnecessarily complicated text processor.
 * After some thought, it seems half a dozen simple nroff macros
 * could perform 90% of the work,
 * and these can be implemented in some 200 lines of code.
 * So I wrote this mini mmx routine that reads an input file,
 * processes it, and produces an output file,
 * ready to cat into an epson printer (or just use prn as the ouptup file).
 * If you want fancier output and you have the means,
 * embed other macros and send the output through tek or troff or tbl or pic,
 * or whatever floats your boat.
 * But for most applications, this routine will suffice.
 *
 * The third and fourth parameters specify page length and width.
 * Pass in zeros to invoke the default settings, appropriate for most printers.
 * Mmx won't split words across lines, and it leaves margins at the top
 * and bottom of each page.  Formfeeds keep the page breaks in sync with
 * the paper.  The following macros are supported.
 *
 * .P paragraph break.  Empty line does the same thing.
 * .br line break.  Jump to the next line.
 * .SK skip to the next page.
 * .DS display starts.  Do not reformat.  Print as is until .DE.
 * .DE display ends.  Resume reformatting operations.
 * .ce center the next line of text.
 *
 * \fI italix printing.
 * \fB bold printing.
 * \fU underline printing.
 * \fR return to roman printing.
 */

#include <stdio.h>

static FILE *inf, *outf; /* input and output for reformatting */
static char input[512];
static char output[132+1];
static int length , width;
static int curline, column;
static char display, center;

extern short ZC_lineno;
extern char *ZC_filename;

mini_mmx(infname, outfname, l, w)
char *infname, *outfname; /* file names */
{
FILE *efopen();

curline = column = 0;
display = center = 0;
length = 66;
width = 76;
if(l) length = l;
if(w) width = w;

inf = efopen(infname, "r");
ZC_filename = infname;
ZC_lineno = 0;
outf = efopen(outfname, "w");

/* top margin of first page */
outc('\n');
outc('\n');

reformat();

if(display) eprint(".DS without a closing .DE");
if(column) putc('\n', outf);
putc('\f', outf);
fclose(inf);
fclose(outf);
} /* mini_mmx */

static reformat()
{
int i, t;

/* run loop getting lines until eof */
while(fgets(input, 511, inf)) {
++ZC_lineno;
i = strlen(input) - 1;
if(input[i] == '\n') input[i] = 0;

/* for your convenience, a blank line outside of a static display */
/* acts just like .P; skips a line. */
if(input[0] == 0 && !display) strcpy(input,".P");

if(input[0] != '.') { /* normal line of text */
if(display) {
if(center) {
/* compute actual length of line */
for(i=t=0; input[i]; ++i) {
if(input[i] != '\\') { ++t; continue; }
if(input[i+1] != 'f') continue;
--t;
}
t = width - t;
if(t < 0) t = 0;
t = t / 2;
for(i=0; i<t; ++i) outc(' ');
} /* centering */

for(i=0; input[i]; ++i)
outc(input[i]);
outc('\n');
if(center) center = display = 0;
continue;
} /* end of static display */

/* print line in normal text mode */
t = 0;
while(input[t]) {
char c = input[t];
if(c == '\t') c = ' ';
if(c == ' ') { outc(c); ++t; continue; }

for(i=0; column+i<width; ++i)
if(input[t+i] == ' ' || input[t+i] == 0) break;
if(column + i >= width)
outc('\n'); /* no room on current line */

for(i=0; column+i<width; ++i) {
if(input[t+i] == ' ' || input[t+i] == 0) break;
outc(input[t+i]);
}
t += i;
} /* input[t] not null */

outc(' ');
/* assumes a double space between sentences */
if(t && (input[t-1] == '.' || input[t-1] == '?' || input[t-1] == '!'))
outc(' ');
continue; /* next line */
} /* end normal text */


/* mmx directives here */

/* .SK, skip page */
if(input[1] == 'S' && input[2] == 'K') {
curline = length;
outc('\n');
continue;
}

/* .DS, static display */
if(input[1] == 'D' && input[2] == 'S') {
if(display) eprint("nested .DS .DE sequences");
display = 1;
if(column) outc('\n');
if(curline != 2) outc('\n');
continue;
}

/* .DE, end static display */
if(input[1] == 'D' && input[2] == 'E') {
if(!display) eprint(".DE with no preceeding .DS");
display = 0;
if(curline != 2) outc('\n');
column = 0;
continue;
}

if(center) eprint("missing line of text after .ce");
if(display) eprint("formatting directives inside .DS .DE block");

/* .P, paragraph */
if(input[1] == 'P') {
if(column) outc('\n');
if(curline != 2) outc('\n');
continue;
}

/* .ce, center line */
if(input[1] == 'c' && input[2] == 'e') {
display = 1;
center = 1;
continue;
}

/* .br, break line */
if(input[1] == 'b' && input[2] == 'r') {
if(column) outc('\n');
continue;
}

eprint("bad mmx directive");
} /* end loop reading lines of the file */
} /* reformat */

/* send character to the printer or screen */
static outc(c)
char c;
{
static esc = 0; /* signal escape sequence */
static font = 0; /* font change */
static spaces = 0;

if(font) {
font = 0;
switch(c) {
static char undo;
case 'P':
case 'R': c = undo; break;
case 'I': c = '4'; undo = '5'; break;
case 'B': c = '3'; undo = '6'; break;
case 'U': c = '-'; undo = '-'; break;
default: eprint("bad \\fX font directive");
} /* switch */
putc('\033', outf); putc(c, outf);
return;
}

if(esc && c == 'f')
{ /* font directive */
esc = 0;
font = 1;
return;
}

if(c == '\7') { putc(c, outf); return; }

if(c == '\n')
{
spaces = 0;
column = 0;
putc(c, outf);
if(++curline >= length-2)
{
fprintf(outf, "\f\n\n");
curline = 2;
}
return;
}

if(c < ' ' && c != '\t') return; /* unprintable */
if(c <= ' ' && !display && !column) return;
if(c == ' ') { ++spaces; ++column; return; }

while(spaces) { putc(' ', outf); --spaces; }

if(c == '\\')
{
if(esc)
{
esc = 0;
putc(c, outf);
++column;
return;
}
esc = 1;
return;
}

esc = 0;

putc(c, outf);
/* column is meaningless in .DS.DE blocks */
/* since that is the only time we output tabs */
/* don't worry about tab column computations */
++column;
} /* outc */

